/****************** <lalVerbatim file="BlockHV">
Author: Tibbits, M. M.
$Id: Block.h,v 1.4 2003/01/22 23:06:50 tibbits Exp $
********************************* </lalVerbatim> */

/******************************* <lalLaTeX file="BlockH">
\section{Header \texttt{Block.h}}
\label{s:Block.h}

Provides structures and definitions global to the block search,
notably \texttt{BlockSearchParams}.

\subsection*{Error conditions}

\input{BlockHErrTab}

\subsection*{Structures and Unions}

*************************************************** </lalLaTeX> */

#ifndef _BLOCK_H_
#define _BLOCK_H_

#include <math.h>
#include <mpi.h>
#include <lal/LALRCSID.h>
#include <LALWrapperInterface.h>
#include <lal/LALStdlib.h>
#include <lal/AVFactories.h>
#include <lal/SeqFactories.h>
#include <lal/LALConstants.h>
#include <lal/LALMoment.h>
#include <lal/Comm.h>
#include "Matrix.h"

/* C++ protection */
#ifdef  __cplusplus
extern "C" {
#endif

NRCSID( BLOCKH, "$Id: Block.h,v 1.4 2003/01/22 23:06:50 tibbits Exp $");


/******** <lalErrTable file="BlockHErrTab"> ********/
#define BLOCKH_ENULL 1
#define BLOCKH_ENNUL 2
#define BLOCKH_EALOC 3
#define BLOCKH_ENUMZ 4
#define	BLOCKH_EARG  5
#define BLOCKH_EDATA 6

#define BLOCKH_MSGENULL "Null pointer"
#define BLOCKH_MSGENNUL "Non-null pointer"
#define BLOCKH_MSGEALOC "Memory allocation error"
#define BLOCKH_MSGENUMZ "Data segment length is zero"
#define	BLOCKH_MSGEARG  "Error parsing command-line arguments"
#define BLOCKH_MSGEDATA "Too few input data points to create useable search data"

/******** </lalErrTable> ********/

/******************************* <lalLaTeX file="BlockH">

\subsubsection*{Union \texttt{ExchType}}
\index{\texttt{ExchType}}

\noindent
The enumerations are:\\
\\
PassData\\
GetData\\
PassResult\\
RePass\\
ExchFinished\\

*************************************************** </lalLaTeX> */

/*  Identifiers for MPI messages types  */
enum
{
	PassData,
	GetData,
	PassResult,
	RePass,
	ExchFinished
}
ExchType;

/******************************* <lalLaTeX file="BlockH">

\subsubsection*{Structure \texttt{BlockSearchParams}}
\index{\texttt{BlockSearchParams}}

\noindent
The fields are:\\
\\

*************************************************** </lalLaTeX> */

/*  Slice of one Level of wavelet decomposition  */
typedef struct
tagwaveletSeg
{
	REAL8Vector		*data;

	INT4			startTime;
	INT4			duration;
	INT2			level;

	REAL8			rho2;
	REAL8			value;
	INT4			index;

	REAL8			variance;
	REAL8			mean;

	struct waveletSeg	*right;
	struct waveletSeg	*left;
	struct waveletSeg	*next;
}
waveletSegment;

/*  Main Data Storage  */
typedef struct
tagWaveletObject
{
	waveletSegment		**decomposition;
	gpsTimeInterval		times;

	INT4			sampleRate;
	INT2			levels;
	INT4			segmentID;

	/* Phase? */
}
waveletObject;

/*  Result Object  */
typedef struct
tagBlockRes
{
	INT2			waveletLevel;
	INT4			segmentID;

	INT4			startTime;
	INT4			duration;

	REAL8			rho2;
	REAL8			value;
	INT4			index;

	REAL8			variance;
	REAL8			mean;

	struct BlockRes		*next;
}
BlockResult;

/*  Main Parameter Structure  */
typedef struct
tagBlockSearchParams
{
	gpsTimeInterval		times;			/*  The GPS start and end time of the data   */

	waveletSegment		*head;			/*  Head of Queue for Breadth First Search  */
	waveletSegment		*tail;			/*  Tail of Queue for Breadth First Search  */
	waveletObject		*data;			/*  Pointer to head of linked list  */

	BlockResult		**results;		/*  Result Objects Received from slave nodes  */
	BlockResult		**prevResults;		/*  Point to end of linked list  */

	dataBase		*resultsToStore;	/*  database results linked list (attached to output obj.)  */
	dataBase		*lastResultStored;	/*  database results linked list ( current->previous )  */

	REAL8			threshold;		/*  Threshold value for recording an event  */
	REAL8			overlap;		/*  Seconds at end of previous sample to include at beginning of next sample  */

	UINT4			numberOfSequences;	/*  Number of Sequences in data  */
	UINT4			segmentsRemaining;	/*  Number of Segments left to receive  */
	UINT4			resultsObtained;	/*  Number of Sequences that the Master has received results for  */
	UINT4			numSlaves;		/*  Number of friends in Comm World  */
	UINT4			numOutput;		/*  Stores number of points in params->resultsToStore  */

	INT4			minVectorLength;	/*  Boolean, To stop or not to stop?  */
	INT4			maxVectorLength;	/*  Parameter to tell the search to chop segments  */
	INT4			numberResults;		/*  Number of Result Objects in Linked List  */
	INT4			searchMaster;		/*  Boolean, am I in charge?  */
	INT4			rank;			/*  node number in Comm World  */

	INT2			overlapSwitch;		/*  Switch to change overlap processing  */

	CHAR			channel[dbNameLimit];	/*  channel of data origin  */
	CHAR			ifo[dbNameLimit];	/*  detector of data origin  */
	CHAR			id[dbNameLimit];	/*  ID tag for easy Database Retrieval  */
}
BlockSearchParams;

/*  Non-LAL Function Prototypes  */

void BlockDistributeDataMaster
(
    LALStatus             *status,
    LALSearchInput        *inout,
    void                  *searchParams,
    LALMPIParams          *mpiParams
);

void BlockDistributeDataSlave
(
    LALStatus             *status,
    LALSearchInput        *inout,
    void                  *searchParams,
    LALMPIParams          *mpiParams
);

void BlockParseData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams
);

void ApplySearchMaster
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*params
);

void ApplySearchSlave
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*params
);

void BuildOutput
(
	LALStatus		*status,
	LALSearchOutput		*output,
	BlockSearchParams	*params
);

void AddResults
(
	LALStatus		*status,
	LALSearchOutput		*output,
	BlockSearchParams	*params
);

/* Necessary Search Tree Functions  */
void BlockAddQ
(
	LALStatus	*status,
	waveletSegment	*newSegment,
	void		*searchParams
);

waveletSegment* BlockDelQ
(
	void		*searchParams
);

void BlockSplitLevel
(
	LALStatus	*status,
	waveletSegment	*input,
	void		*searchParams
);

/* C++ protection */
#ifdef  __cplusplus
}
#endif

/******************************* <lalLaTeX file="BlockH">
  
\vfill{\footnotesize\input{BlockHV}}

\newpage\input{BlockInitSearchC}
\newpage\input{BlockConditionDataC}
\newpage\input{BlockApplySearchC}
\newpage\input{BlockRho2C}
\newpage\input{BlockFinalizeSearchC}

*************************************************** </lalLaTeX> */

#endif /* _BLOCK_H_ */
