/******** <lalVerbatim file="BlockApplySearchCV"> ********
Author: Tibbits, M. M.
$Id: BlockApplySearch.c,v 1.5 2003/02/10 19:17:43 tibbits Exp $
********* </lalVerbatim> ********/

/******************************* <lalLaTeX file="BlockApplySearchC">
\subsection{Module \texttt{BlockApplySearch.c}}
\label{s:BlockApplySearch.c}

Routine that uses parallelism of processes to implement the rho2 backround
search.  The rho2 search is based on, a yet to be published, block
normailization paper.  Our search returns a result or "expectation" value
which corresponds to the probability of there being multiple gaussian
waveforms within the measured data.  The search also returns the index
and magnitude of the maximum value in the input vector.  Via a filterparam
you can also enable the Rho3 search which chops the input at the maximum
value and reruns the Rho2 code on both segements.  Thereby, we can search
for both the beginning and the end of a gaussian waveform present.

\subsubsection*{Prototypes}
\input{BlockApplySearchCP}
\index{\texttt{LALApplySearch()}}

\subsection*{Description}

Within the LALApplySearch Routine, the block search master receives the
first set of results from the slaves (processed within the LALApplySearch
on the slave node).  Then the master node either sends a data id of -1
signalling that all vectors that need processing have been sent out or the
master node send the index of the vector stored within the master's parameter
structure in a vector sequence "data".

On each slave node, as in the LALConditionData section, the slave first
processes the data vector that it received in LALConditionData.  Next the 
slave receives the id of the next vector and either exits if -1 is given or
waits to receive the next data vector if signalled by the master node.  At
the completetion of Rho2, if the Rho3 flag is set, each slave will wait for
the next input vecto to process.

The result values are stored in an array on the master node until all vectors
have been processed.  The results are then built for subsequent insertion
into the metadata database and frames using the \texttt{BlockBuildOutput}
routine.

\subsection*{Algorithms}

The LALApplySearch routine does not manipulate the data in any way on either
master or the slave until the slave sends the data to the Rho2 routine.  The
LALApplySearch routine is designed as a wrapper to the facilitate 
parallelism mentioned earlier.  The Block Normal Search takes an input of a
vector sequence that can have any number of vectors in it.  The parallelism
begins by breaking the input into individual vectors and sending each
vector to a different slave node.  Each slave node then preforms the Rho2
algorithm on its vector and returns the results

To note, the Rho2 algorithm is discussed below under the \texttt{BlockRho2}
routine heading.

The MPI exchange is as follows:
%
\begin{enumerate}
\item 
The Master waits to receive a message from a slave node that the slave
is ready to send its data.

\item 
The Slave then sends first the index of the data vector that it
processed.

\item 
Then the Slave sends its results

\item
The Master then sends back the index of the next vector to be processed
which is set to -1 if all data vectors have been processed.

\item
The Master then sends the next data vector to the slave, given that
there are still vectors to be processed.

\end{enumerate}

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{LALMPISendMsg}
\item \texttt{LALMPIRecvMsg}
\item \texttt{LALMPISendINT4}
\item \texttt{LALMPIRecvINT4}
\item \texttt{LALMPISendREAL8}
\item \texttt{LALMPIRecvREAL8}
\item \texttt{LALDCreateVector}
\item \texttt{LALMPISendREAL8Vector}
\item \texttt{LALMPIRecvREAL8Vector}
\item \texttt{BlockRho2}
\item \texttt{BlockBuildOutput} 
\item \texttt{LALDDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{BlockApplySearchCV}}

 *************************************************** </lalLaTeX> */

#include "Block.h"

NRCSID( BLOCKAPPLYSEARCHC, "$Id: BlockApplySearch.c,v 1.5 2003/02/10 19:17:43 tibbits Exp $");

/************************ <lalVerbatim file="BlockApplySearchCP"> */
void LALApplySearch
(
	LALStatus				*status,
	LALSearchOutput			*output,
	LALSearchInput			*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;

	INITSTATUS( status, "BlockApplySearch", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Inititalize Variables  */
	params = ((BlockSearchParams*)(applyParams->searchParams));

	/*  Sanity Checks on Input Variables  */
	ASSERT( applyParams, status, BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Get rank  */
	MPI_Comm_rank(*(applyParams->mpiParams->comm), &(params->rank));

	/*  Which node is this running on?  */
	if (params->searchMaster)
	{
		/* master */
		ApplySearchMaster(status->statusPtr, output, input, applyParams );
	}
	else
	{
		/* slave */
		ApplySearchSlave(status->statusPtr, output, input, applyParams );
	}

	DETATCHSTATUSPTR (status);
	RETURN(status);
}

/*
 * Function ApplySearchMaster
 *
 * Called from LALApplySearch only on the search master node. It's
 * main job is to wait for responses for each slave ie. retrieve the
 * statistics, send the next vector, wait for the slaves to end, and
 * then to format the output.
 */

/************************ <lalVerbatim file="BlockApplySearchMasterCP"> */
void ApplySearchMaster
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams		*params;
	waveletSegment			*tempPointer;
	waveletObject			*wavelet;
	BlockResult			*current;
	REAL8Vector			*tempVector;
	MPIMessage			mpiMsg;
	MPI_Comm			*comm;
	UINT4				numberResults;
	UINT4				iterator;
	UINT4				level;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataMaster", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Sanity Checks on Variables  */
	ASSERT( input, status, BLOCKH_ENULL, BLOCKH_MSGENULL);
	ASSERT( output, status, BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Variable Initialization  */
	comm	= ((MPI_Comm*)(applyParams->mpiParams->comm));
	params	= ((BlockSearchParams*)(applyParams->searchParams));
	current	= NULL;

	wavelet	= params->data;

	/*  Wait for Slaves to say, "I'm here"  */
	LALMPIRecvMsg( status->statusPtr, &mpiMsg, (*comm));
	CHECKSTATUSPTR( status );

	/*  Slave sends number of results */
	LALMPIRecvINT4(status->statusPtr, &(numberResults), mpiMsg.source, (*comm));
	CHECKSTATUSPTR( status );
#ifdef BLOCKSEARCH__DEBUG
	printf("\nNumber of Results to be sent = %d\n",numberResults);
#endif
	for ( iterator = 0; iterator < numberResults; iterator++)
	{
		current = (BlockResult*)(LALMalloc ( sizeof(BlockResult)));
		if( !current )
		{
			ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
		}

		/*  Slave sends waveletLevel  */
		LALMPIRecvINT2(status->statusPtr, &(current->waveletLevel), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends segment ID  */
		LALMPIRecvINT4(status->statusPtr, &(current->segmentID), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends startTime  */
		LALMPIRecvINT4(status->statusPtr, &(current->startTime), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends duration  */
		LALMPIRecvINT4(status->statusPtr, &(current->duration), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends result  */
		LALMPIRecvREAL8( status->statusPtr, &(current->rho2), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends max value  */
		LALMPIRecvREAL8( status->statusPtr, &(current->value), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends index of max value  */
		LALMPIRecvINT4( status->statusPtr, &(current->index), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		current->next = NULL;

		if( params->results[current->waveletLevel] == NULL )
		{
			params->results[current->waveletLevel] = current;
		}
		else
		{
			params->prevResults[current->waveletLevel]->next = current;
		}

		params->prevResults[current->waveletLevel] = current;
		current = NULL;
	}

	if( numberResults )
	{
		/*  Chain results just received onto stored results  */
		AddResults(status->statusPtr, output, params);
	}

	/*  This counts number of results still to get  */
	params->resultsObtained++;

	if( params->segmentsRemaining == 0 )
	{
		iterator = -1;

		/*  Send kill code  */
		LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
	}
	else	/*  Send next Wavelet Object  */
	{
		/*  Send wavelet->segmentID */
		iterator = (params->numberOfSequences - params->segmentsRemaining) - 1;
		LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send wavelet->levels  */
		LALMPISendINT2( status->statusPtr, &(wavelet->levels), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send decomposition[0]->data->length  */
		LALMPISendINT4( status->statusPtr, &(wavelet->decomposition[0]->data->length), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send startTime  */
		LALMPISendINT4( status->statusPtr, &(wavelet->decomposition[0]->startTime), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send duration  */
		LALMPISendINT4( status->statusPtr, &(wavelet->decomposition[0]->duration), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send Data in for loop  */
		for(level = 0; level < wavelet->levels; level++ )
		{
			tempPointer	= NULL;
			tempVector	= NULL;

			tempPointer = wavelet->decomposition[level];

			LALDCreateVector(status->statusPtr, &tempVector, wavelet->decomposition[level]->data->length);
			CHECKSTATUSPTR( status );

			for( iterator = 0; iterator < tempVector->length; iterator++ )
			{
				tempVector->data[iterator] = wavelet->decomposition[level]->data->data[iterator];
			}

			/*  Send Data to Slave  */
			LALMPISendREAL8Vector(status->statusPtr, tempVector, mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Clean up temporary Mem each time used  */
			LALDDestroyVector(status->statusPtr, &tempVector);
			CHECKSTATUSPTR( status );

			/*  Remove wavelet Segment  */
			wavelet->decomposition[level] = wavelet->decomposition[level]->next;
			LALDDestroyVector(status->statusPtr, &tempPointer->data);
			CHECKSTATUSPTR( status );
			LALFree(tempPointer);

		}	/*  For loop to transmit data  */

		output->fracRemaining	= ( params->segmentsRemaining / params->numberOfSequences );
		params->segmentsRemaining--;

	}		/*  ( params->segmentsRemaining != 0 )  */

	if( params->resultsObtained == params->numberOfSequences )
	{
		/*  Format the result array into an output object to fill the database  */
		BuildOutput( status->statusPtr, output, params );
		output->fracRemaining	= 0;
		output->notFinished	= 0;
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}


/*
 * Function ApplySearchSlave
 *
 * Called from LALApplySearch on each slave node. It receives the data
 * sent by the master node, processes it via the BlockRho2 function,
 * then it passes the results to the master node, and if there are more
 * vectors to be processed, it waits to receive the next one.
 */


/************************ <lalVerbatim file="BlockApplySearchMasterCP"> */
void ApplySearchSlave
(
	LALStatus				*status,
	LALSearchOutput			*output,
	LALSearchInput			*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;
	waveletSegment		*current;
	waveletSegment		*tmp;
	waveletObject		*wavelet;
	BlockResult		*tempResult;

	MPIMessage		sendMsg;
	MPI_Comm		*comm;

	INT4			startTime;
	INT4			duration;
	INT4			iterator;
	INT4			length;
	INT4			level;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataSlave", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	comm	= ((MPI_Comm*)(applyParams->mpiParams->comm));
	params	= ((BlockSearchParams*)(applyParams->searchParams));
	wavelet	= params->data;

	/*  Block Normal Analysis of Data  */
	for( iterator = 0; iterator < params->data->levels; iterator++)
	{
		BlockAddQ( status->statusPtr, params->data->decomposition[iterator], params );
		params->data->decomposition[iterator] = NULL;
	}

	while(( current = BlockDelQ ( params ) ) != NULL )
	{
		/*  Send current data vector myParams->vec to BlockRho2 Function  */
		BlockRho2(status->statusPtr, &current->rho2, &current->value, &current->index, current->data);
		CHECKSTATUSPTR( status );
#ifdef BLOCKSEARCH__DEBUG
		/*  Print a listing of the results  */
		printf("\nwavelet[%d,%d]->result = %3.16f\n", current->level, params->rank, (REAL4)(log(current->rho2)));
		printf("wavelet[%d,%d]->rpeak = %3.16f\n", current->level, params->rank, (REAL4)(log(current->value)));
		printf("wavelet[%d,%d]->index = %d\n", current->level, params->rank, (current->index));
#endif
		BlockSplitLevel( status->statusPtr, current, params );
		CHECKSTATUSPTR( status );

		if ( current->left != NULL )
		{
			tmp = current->left;
			BlockAddQ ( status->statusPtr, tmp, params );
			CHECKSTATUSPTR( status );
		}

		if ( current->right != NULL )
		{
			tmp = current->right;
			BlockAddQ ( status->statusPtr, tmp, params );
			CHECKSTATUSPTR( status );
		}

		LALDDestroyVector( status->statusPtr, &current->data );
		LALFree( current );
	}

	/*  Set up MPI Message  */
	sendMsg.source	= params->rank;
	sendMsg.msg	= PassData;
	sendMsg.send	= 1;

	/*  Send Message to say I am ready  */
	LALMPISendMsg( status->statusPtr, &sendMsg, 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Slave sends number of results */
	LALMPISendINT4(status->statusPtr, &(params->numberResults), 0, (*comm));
	CHECKSTATUSPTR( status );

	for ( iterator = 0; iterator < params->numberResults; iterator++)
	{
		level = 0;

		while( params->results[level] == NULL )
			level++;

		tempResult = params->results[level];

		params->results[level] = params->results[level]->next;

		/*  Slave sends waveletLevel  */
		LALMPISendINT2(status->statusPtr, &(tempResult->waveletLevel), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends segment ID  */
		LALMPISendINT4(status->statusPtr, &(tempResult->segmentID), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends startTime  */
		LALMPISendINT4(status->statusPtr, &(tempResult->startTime), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends duration  */
		LALMPISendINT4(status->statusPtr, &(tempResult->duration), 0, (*comm));

		/*  Slave sends result  */
		LALMPISendREAL8( status->statusPtr, &(tempResult->rho2), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends max value  */
		LALMPISendREAL8( status->statusPtr, &(tempResult->value), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends index of max value  */
		LALMPISendINT4( status->statusPtr, &(tempResult->index), 0, (*comm));
		CHECKSTATUSPTR( status );

		LALFree( tempResult );
		CHECKSTATUSPTR(status);
	}

	params->head = params->tail = NULL;
	current = NULL;

	LALFree( params->data->decomposition );
        LALFree( params->results );
	LALFree( params->prevResults );

	params->prevResults = NULL;

	/*  Here we have completely processed a sequence and returned the results  */
	/*  Now we send a message to ask if there is more data to work on  */

	/*  Receive wavelet->segmentID */
	LALMPIRecvINT4( status->statusPtr, &(wavelet->segmentID), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Are we done??  */
	if( wavelet->segmentID < 0 )
	{
		LALFree( params->data );
		CHECKSTATUSPTR(status);

		params->data = NULL;

		/*  This halts ApplySearch on the slave node  */
		output->fracRemaining	= 0;
 		output->notFinished 	= 0;

		DETATCHSTATUSPTR( status );
		RETURN( status );
	}

	
	/*  Here we begin processing the next sequence  */
	/*  Receive wavelet->levels  */
	params->numberResults = 0;
	LALMPIRecvINT2( status->statusPtr, &(wavelet->levels), 0, (*comm));
	CHECKSTATUSPTR( status );

	wavelet->decomposition = (waveletSegment**)(LALMalloc ( sizeof(waveletSegment*) * wavelet->levels ));
	if( !wavelet->decomposition)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

        params->results = (BlockResult**)(LALMalloc ( sizeof(BlockResult*) * wavelet->levels ));
	if( !params->results)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	params->prevResults = (BlockResult**)(LALMalloc ( sizeof(BlockResult*) * wavelet->levels ));
	if( !params->prevResults )
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	for( level = 0; level < wavelet->levels; level++)
	{
		params->results[level] = NULL;
		params->prevResults[level] = NULL;
	}

	/*  Receive decomposition[0]->data->length  */
	LALMPIRecvINT4( status->statusPtr, &(length), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive startTime  */
	LALMPIRecvINT4( status->statusPtr, &(startTime), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive duration  */
	LALMPIRecvINT4( status->statusPtr, &(duration), 0, (*comm));
	CHECKSTATUSPTR( status );

	for( level = 0; level < wavelet->levels; level++)
	{
		wavelet->decomposition[level] = current = (waveletSegment*)( LALMalloc( sizeof(waveletSegment)));

		current->rho2 = current->index = current->value = 0;
		current->left	= NULL;
		current->right	= NULL;
		current->data	= NULL;
		current->next	= NULL;

		current->startTime = startTime;
		current->duration = duration;

		current->level = level;

		LALDCreateVector( status->statusPtr, &(current->data), length);

		/*  Receive data from Master  */
		LALMPIRecvREAL8Vector(status->statusPtr, current->data, 0, (*comm));
		CHECKSTATUSPTR( status );

		if ( level + 1 < wavelet->levels )
		{
			length /= 2;
		}
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}
