/******** <lalVerbatim file="BlockConditionDataCV"> ********
Author: Tibbits, M. M.
$Id: BlockConditionData.c,v 1.5 2003/02/10 19:17:43 tibbits Exp $
********* </lalVerbatim> ********/

/******************************* <lalLaTeX file="BlockConditionDataC">
\subsection{Module \texttt{BlockConditionData.c}}
\label{s:BlockConditionData.c}

Routine that processes the input data object and repacks it into the
block parameter structure.  It then also sends out the first vector
to be processed.

\subsubsection*{Prototypes}
\input{BlockConditionDataCP}
\input{BlockDistributeDataMasterCP}
\input{BlockDistributeDataSlaveCP}
\input{BlockParseDataCP}
\index{\texttt{LALConditionData()}}

\subsection*{Description}

Within the \texttt{LALConditionData} Routine, the block search master sends
the input object to the \texttt{BlockParseData} function.  This function
repacks the input in the masters parameter structure and stores the input
data in the parameter's data member.  The Master then sends the first vector
to each node to begin processing.  On each slave node, each slave simply
receives the first data vector and then exits.

\subsection*{Algorithms}

At this time, the \texttt{LALConditionData} routine does not manipulate
the data in any way.  It merely repacks the data in a LAL Datatype.

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{BlockDistributeDataMaster}
\item \texttt{BlockDistributeDataSlave}
\item \texttt{LALMPISendMsg}
\item \texttt{LALMPIRecvMsg}
\item \texttt{LALMPISendINT4}
\item \texttt{LALMPIRecvINT4}
\item \texttt{LALMPISendREAL8}
\item \texttt{LALMPIRecvREAL8}
\item \texttt{LALDCreateVector}
\item \texttt{LALMPISendREAL8Vector}
\item \texttt{LALMPIRecvREAL8Vector}
\item \texttt{BlockParseData}
\item \texttt{LALDCreateVectorSequence}
\item \texttt{BlockBuildOutput} 
\item \texttt{LALDDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{BlockConditionDataCV}}

 *************************************************** </lalLaTeX> */

#include "Block.h"

NRCSID( BLOCKCONDITIONDATAC, "$Id: BlockConditionData.c,v 1.5 2003/02/10 19:17:43 tibbits Exp $");

/************************ <lalVerbatim file="BlockConditionDataCP"> */
void LALConditionData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*myParams;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "ConditionData", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Sanity Checks on Paramter Structure  */
	ASSERT(searchParams, status,BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Variable Initialization  */
	myParams = ((BlockSearchParams*)(searchParams));

	/*  Get rank  */
	MPI_Comm_rank(*(mpiParams->comm), &(myParams->rank));

	/*  Start Master/Slave version  */
	/*  Which node is this running on?  */
	if (myParams->searchMaster)
	{
		/* master */
		BlockDistributeDataMaster(status->statusPtr, inout, searchParams, mpiParams );
	}
	else
	{
		/* slave */
		BlockDistributeDataSlave(status->statusPtr, inout, searchParams, mpiParams );
	}

	DETATCHSTATUSPTR( status );
	RETURN( status );
}


/*
 * Function BlockDistributeDataMaster
 *
 * Called from LALConditionData only on the search master node. It
 * uses the BlockParseData routine to repack the input and then it
 * sends the a data vector to each waiting slave.
 */

/************************ <lalVerbatim file="BlockDistributeDataMasterCP"> */
void BlockDistributeDataMaster
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;
	waveletSegment		*tempPointer;
	waveletObject		*wavelet;
	REAL8Vector		*tempVector;
	MPIMessage		mpiMsg;
	MPI_Comm		*comm;
	UINT4			currSlaves;
	UINT4			iterator;
	UINT4			level;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataMaster", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Check that input exists  */
	ASSERT(inout, status, BLOCKH_ENULL, BLOCKH_MSGENULL);
	ASSERT(inout->sequences, status, BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Variable Initialization  */
	comm	= ((MPI_Comm*)(mpiParams->comm));
	params	= ((BlockSearchParams*)(searchParams));

	/* get the number of search slaves */
	MPI_Comm_size(*(mpiParams->comm), &(params->numSlaves));

	/* The number returned INCLUDES the search master, so subtract 1 */
	params->numSlaves -= 1;

	currSlaves	= params->numSlaves;

	/*  Repack data into usable LAL Structure  */
	BlockParseData ( status->statusPtr, inout, params );

	wavelet = params->data;

	/*  While there are still Slaves without Data, send more  */
	while(currSlaves > 0 )
	{
		/*  Wait for Slaves to say, "I'm here"  */
		LALMPIRecvMsg( status->statusPtr, &mpiMsg, (*comm));
		CHECKSTATUSPTR( status );

		if( params->segmentsRemaining == 0 )
		{
			iterator = -1;

			/*  Send kill code  */
			LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );
			continue;
		}
		else
		{
			params->segmentsRemaining--;
		}

		/*  Send wavelet->segmentID */
		iterator = (params->numberOfSequences - params->segmentsRemaining) - 1;

		LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send wavelet->levels  */
		LALMPISendINT2( status->statusPtr, &(wavelet->levels), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send decomposition[0]->data->length  */
		LALMPISendINT4( status->statusPtr, &(wavelet->decomposition[0]->data->length), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send startTime  */
		LALMPISendINT4( status->statusPtr, &(wavelet->decomposition[0]->startTime), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send duration  */
		LALMPISendINT4( status->statusPtr, &(wavelet->decomposition[0]->duration), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send Data in for loop  */
		for(level = 0; level < wavelet->levels; level++ )
		{
			tempPointer	= NULL;
			tempVector	= NULL;

			LALDCreateVector(status->statusPtr, &tempVector, wavelet->decomposition[level]->data->length);
			CHECKSTATUSPTR( status );

			for( iterator = 0; iterator < tempVector->length; iterator++ )
			{
				tempVector->data[iterator] = wavelet->decomposition[level]->data->data[iterator];
			}

			/*  Send Data to Slave  */
			LALMPISendREAL8Vector(status->statusPtr, tempVector, mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Clean up temporary Mem each time used  */
			LALDDestroyVector(status->statusPtr, &tempVector);
			CHECKSTATUSPTR( status );

			tempPointer = wavelet->decomposition[level];
			wavelet->decomposition[level] = wavelet->decomposition[level]->next;

			/*  Remove Sent Wavelet Data  */
			LALDDestroyVector(status->statusPtr, &tempPointer->data);
			LALFree(tempPointer);
		}

		currSlaves--;
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}

/*
 * Function BlockDistributeDataSlave
 *
 * Called from LALConditionData on each slave node. It
 * waits for the master to send the first data vector for
 * it to process and then it exits.
 */

/************************ <lalVerbatim file="BlockDistributeDataSlaveCP"> */
void BlockDistributeDataSlave
(
	LALStatus		*status,
	LALSearchInput	*inout,
	void			*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;
	waveletSegment		*current;
	waveletObject		*wavelet;
	MPIMessage		sendMsg;
	MPI_Comm		*comm;
	INT4			startTime;
	INT4			duration;
	INT4			length;
	INT4			level;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataSlave", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	params	= ((BlockSearchParams*)(searchParams));
	comm	= ((MPI_Comm*)(mpiParams->comm));

	params->data = wavelet = (waveletObject*)(LALMalloc( sizeof( waveletObject)));
	if(!wavelet)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	/*  Set up MPI Message  */
	sendMsg.source	= params->rank;
	sendMsg.msg	= PassData;
	sendMsg.send	= 1;

	/*  Send Message to say I am ready  */
	LALMPISendMsg( status->statusPtr, &sendMsg, 0, (*comm));
	CHECKSTATUSPTR( status );	

	/*  Receive wavelet->segmentID */
	LALMPIRecvINT4( status->statusPtr, &(wavelet->segmentID), 0, (*comm));
	CHECKSTATUSPTR( status );

	if( wavelet->segmentID < 0 )
	{
		DETATCHSTATUSPTR( status );
		RETURN( status );
	}

	/*  Receive wavelet->levels  */
	LALMPIRecvINT2( status->statusPtr, &(wavelet->levels), 0, (*comm));
	CHECKSTATUSPTR( status );

	wavelet->decomposition = (waveletSegment**)(LALMalloc ( sizeof(waveletSegment*) * wavelet->levels ));
	if( !wavelet->decomposition)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	params->results = (BlockResult**)(LALMalloc ( sizeof(BlockResult*) * wavelet->levels ));
	if( !params->results )
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	params->prevResults = (BlockResult**)(LALMalloc ( sizeof(BlockResult*) * wavelet->levels ));
	if( ! params->prevResults )
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	for( level = 0; level < wavelet->levels; level++)
	{
		params->results[level] = NULL;
		params->prevResults[level] = NULL;
	}

	params->numberResults = 0;

	/*  Receive decomposition[0]->data->length  */
	LALMPIRecvINT4( status->statusPtr, &(length), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive startTime  */
	LALMPIRecvINT4( status->statusPtr, &(startTime), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive duration  */
	LALMPIRecvINT4( status->statusPtr, &(duration), 0, (*comm));
	CHECKSTATUSPTR( status );

	for( level = 0; level < wavelet->levels; level++)
	{
		wavelet->decomposition[level] = current = (waveletSegment*)( LALMalloc( sizeof(waveletSegment)));

		current->rho2 = current->index = current->value = 0;
		current->left	= NULL;
		current->right	= NULL;
		current->data	= NULL; 
		current->next	= NULL;

		current->startTime = startTime;
		current->duration = duration;
		current->level = level;

		LALDCreateVector( status->statusPtr, &(current->data), length);
		CHECKSTATUSPTR( status);

		/*  Send Data to Slave  */
		LALMPIRecvREAL8Vector(status->statusPtr, current->data, 0, (*comm));
		CHECKSTATUSPTR( status );

		if ( level + 1 < wavelet->levels )
		{
			length /= 2;
		}
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}

/************************ <lalVerbatim file="BlockParseDataCP"> */
void BlockParseData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams
)
/*********************************************** </lalVerbatim> */
{

	/*  Variable Declarations  */
	BlockSearchParams	*params;
	waveletSegment		*current;
	waveletObject		*wavelet;
	UINT4			iterator;
	UINT4			level;
	UINT4			offset;
	INT4			sampleRate;
	INT4			length;
	INT4			index;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "BlockParseData", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	params	= ((BlockSearchParams*)(searchParams));

	/*  Initialize the array of Pointers  */
	wavelet	= (waveletObject*)(LALMalloc( sizeof(waveletObject) ));
	if( !wavelet )
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}


	length = inout->sequences[0].dimensions[0]/params->maxVectorLength;
	if((inout->sequences[0].dimensions[0] - length*params->maxVectorLength) >= params->minVectorLength)
	{
		length++;
	}
#ifdef BLOCKSEARCH__DEBUG
	printf("\nparams->max = %d  total / max = %d\n\n", params->maxVectorLength, length);
#endif
	params->segmentsRemaining = params->numberOfSequences = length;

	/*  Assert Usable amount of Data  */
	if ( length < 2 )
	{
		ABORT(status, BLOCKH_EDATA, BLOCKH_MSGEDATA);
	}

	wavelet->levels = inout->numberSequences;
#ifdef BLOCKSEARCH__DEBUG
	printf("\nInput wavelet decompositions = %d \n",wavelet->levels);
#endif
	/*  Outer loop steps through wavelet levels  */
	/*  Middle loop steps through wavelet segments  */
	/*  Inner loop copies wavelet data  */
    
	wavelet->decomposition = (waveletSegment**)(LALMalloc ( sizeof(waveletSegment*) * wavelet->levels ));
	if( !wavelet->decomposition)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}
	
	params->results = (BlockResult**)(LALMalloc ( sizeof(BlockResult*) * wavelet->levels ));
	if( !params->results)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	params->prevResults = (BlockResult**)(LALMalloc ( sizeof(BlockResult*) * wavelet->levels ));
	if( !params->prevResults )
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	sampleRate = params->maxVectorLength;

	for( level = 0; level < wavelet->levels; level++)
	{
		params->results[level] = NULL;
		params->prevResults[level] = NULL;

		wavelet->decomposition[level] = current = (waveletSegment*)( LALMalloc( sizeof(waveletSegment)));
#ifdef BLOCKSEARCH__DEBUG
		printf("\nMSG: inout->sequences[%d].dimensions[0] = %d \n\n",level,inout->sequences[level].dimensions[0]);
#endif
		for( iterator = 0; iterator < params->numberOfSequences - 1; iterator++)
		{
			offset 	= iterator * sampleRate;

			current->rho2	= current->index = current->value = 0;

			current->left	= NULL;
			current->right	= NULL;
			current->data	= NULL;

			LALDCreateVector( status->statusPtr, &(current->data), sampleRate);
			CHECKSTATUSPTR( status );

			for( index = 0; index < sampleRate; index++ )
			{
				current->data->data[index] = (REAL8)(inout->sequences[level].data.real4[index + offset]);
			}

			/*  Must correct for metadata  */
			current->startTime	= offset * ( params->maxVectorLength / sampleRate );

			/*  Regardless of level, duration of full segment will always = Max  */
			current->duration	= params->maxVectorLength;

			current->level = level;

			current->next = ((waveletSegment*)(LALMalloc( sizeof(waveletSegment))));
			if( !current->next )
			{
				ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
			}

			current = current->next;
		}

		offset = (params->numberOfSequences - 1)*sampleRate;

		current->rho2 = current->index = current->value = 0;

		current->left	= NULL;
		current->right	= NULL;
		current->data	= NULL;
		current->next	= NULL;

		/*  Must correct for metadata  */
		current->startTime = offset * ( params->maxVectorLength / sampleRate );

		current->level = level;

		length = inout->sequences[level].dimensions[0] - (params->numberOfSequences - 1) * sampleRate;

		current->duration = length;
#ifdef BLOCKSEARCH__DEBUG
		printf("\nMSG:Length of last vector := %d \n",length);
#endif
		current->data = NULL;
		LALDCreateVector( status->statusPtr, &(current->data), length);
		CHECKSTATUSPTR( status );

		for( index = offset; index < inout->sequences[level].dimensions[0]; index++ )
		{
			current->data->data[index - offset]=(REAL8)(inout->sequences[level].data.real4[index]);
		}

		if ( level + 1 < wavelet->levels )
		{
			sampleRate /= 2;
		}
	}

	/*  Global metadata  */
	params->times.startSec = (INT8) inout->sequences[0].range.dTime.startSec;
	params->times.startNan = (INT8) inout->sequences[0].range.dTime.startNan;
	params->times.stopSec = (INT8) inout->sequences[0].range.dTime.stopSec;
	params->times.stopNan = (INT8) inout->sequences[0].range.dTime.stopNan;
	params->times.timeStepSize = (INT8) inout->sequences[0].range.dTime.timeStepSize;
	params->times.numberSamples = (INT8) inout->sequences[0].range.dTime.numberSamples;

	/*  This is not correct for production model  */
	/*  HARDCODED Channel Name  */
	strcpy(params->channel, "H2");
	strcpy(params->ifo, "H2");
	
	/*  Store newly created wavelet object in master's params structure  */
	params->data = wavelet;

	/*  Normal Exit  */
	DETATCHSTATUSPTR (status);
	RETURN(status);
}
