/*----------------------------------------------------------------------- 
 * 
 * File Name: ConditionData.c
 *
 * Author: Brady, P. R. and Brown, D. A.
 * 
 * Revision: $Id: ConditionData.c,v 1.66 2003/02/07 20:21:30 duncan Exp $
 * 
 *-----------------------------------------------------------------------
 */

#include "ConditionData.h"


/*
 *
 * prototypes for debugging functions
 *
 */


#ifdef INSPIRAL_SO_DEBUG
static void
PrintInspiralBank (
    LALStatus           *status,
    InspiralTemplate    *head,
    FILE                *fp
    );

static void
PrintTemplateNodeList(
    LALStatus                  *status,
    InspiralTemplateNode       *tmpltHead,
    FILE                       *fp
    );
#endif


NRCSID( CONDITIONDATAC, PKG "$Id: ConditionData.c,v 1.66 2003/02/07 20:21:30 duncan Exp $" );

/*-----------------------------------------------------------------------*/
void
LALConditionData(
    LALStatus             *status,
    LALSearchInput        *inout,
    void                  *searchParams,
    LALMPIParams          *mpiParams
    )
{
  UINT4                         i;
  UINT4                         commSize;
  UINT4                         commRank;
  UINT4                         haveChan = 0;
  UINT4                         haveSpec = 0;
  UINT4                         haveResp = 0;
  UINT4                         calIdx = 0;
  CHAR                          ifo[3];
  UINT4                         haveNewCal    = 0;
  INT8                          chanStartTime = 0;
  INT8                          chanStopTime  = 0;
  LIGOTimeGPS                   chanStopEpoch;
  LIGOTimeGPS                   responseEpoch;
  REAL4                        *dataPtr         = NULL;
  multiDimData                 *mddTmplt        = NULL;
  multiDimData                 *mddInjEpochs    = NULL;
  multiDimData                 *mddInjParams    = NULL;
  InspiralSearchParams         *params          = NULL;
  InspiralTemplate             *currentTmplt    = NULL;
  InspiralTemplateNode         *tmpltNode       = NULL;
  REAL4TimeSeries              *chanPtr         = NULL;
  REAL4FrequencySeries         *specPtr         = NULL;
  COMPLEX8FrequencySeries      *respPtr         = NULL;
  const LALUnit strainPerCount = {0,{0,0,0,0,0,1,-1},{0,0,0,0,0,0,0}};


  INITSTATUS( status, "LALConditionData", CONDITIONDATAC );
  ATTATCHSTATUSPTR( status );


 /* 
  *
  * check and cast parameter and input structures
  *
  */


  ASSERT( searchParams, status,
      CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );

  params = (InspiralSearchParams *) searchParams;

  chanPtr = &(params->chan);
  specPtr = &(params->spec);
  respPtr = &(params->resp);

  ASSERT (inout, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL);

  MPI_Comm_rank( *(mpiParams->comm), &commRank );
  MPI_Comm_size( *(mpiParams->comm), &commSize );

  if ( params->searchMaster )
  {


   /*
    *
    * get the necessary information from the input ilwd data
    *
    */


    for ( i = 0; i < inout->numberSequences; ++i )
    {
      if ( strstr( inout->sequences[i].name, INPUTNAME_CHANNEL ) )
      {

        
        /*
         *
         * construct the search summary information from the input data 
         *
         */


        INT8    outputStartNS;
        INT8    outputEndNS;
        REAL8   chanDeltaT = inout->sequences[i].range.dTime.timeStepSize;
        INT8    tIgnoreNS = (INT8) 
          ( 1e9 * (REAL8) ( params->initParams.numPoints / 4 ) * chanDeltaT );

        /* record the channel name */
        if ( ! inout->sequences[i].range.dTime.channelName )
        {
          ABORT( status, CONDITIONDATAH_ENULL, 
              CONDITIONDATAH_MSGENULL " channelName" );
        }

        strncpy( params->chan.name,
            inout->sequences[i].range.dTime.channelName[0], 
            LALNameLength * sizeof(CHAR) );

        /* record the start and stop times of the input data */
        params->info.inputStartTime.gpsSeconds = 
          inout->sequences[i].range.dTime.startSec;
        params->info.inputStartTime.gpsNanoSeconds = 
          inout->sequences[i].range.dTime.startNan;
        params->info.inputEndTime.gpsSeconds = 
          inout->sequences[i].range.dTime.stopSec;
        params->info.inputEndTime.gpsNanoSeconds = 
          inout->sequences[i].range.dTime.stopNan;

        /* XXX compute and store the time actually searched for chirps      */
        /* XXX this must match the hardwired parameter in FindChirpFilter.c */
        /* XXX around line 639 (search for MSGECRUP)                        */
        /* XXX we ignore 1/4 segment at the start and end of the input data */
        LALGPStoINT8( status->statusPtr, 
            &outputStartNS, &(params->info.inputStartTime) );
        CHECKSTATUSPTR( status );
        LALGPStoINT8( status->statusPtr, 
            &outputEndNS, &(params->info.inputEndTime) );
        CHECKSTATUSPTR( status );

        outputStartNS += tIgnoreNS;
        outputEndNS   -= tIgnoreNS;

        LALINT8toGPS( status->statusPtr, 
            &(params->info.outputStartTime), &outputStartNS );
        CHECKSTATUSPTR( status );
        LALINT8toGPS( status->statusPtr, 
            &(params->info.outputEndTime), &outputEndNS );
        CHECKSTATUSPTR( status );
      }
      else if ( strstr( inout->sequences[i].comment, INPUTNAME_TMPLTBANK ) ||
          strstr( inout->sequences[i].name, INPUTNAME_TMPLTBANK ) )
      {

        
        /*
         *
         * save the location of the template bank 
         *
         */


        mddTmplt = &( inout->sequences[i] );
      }
    }


    /*
     *
     * construct the template bank from the input ilwd mutiDimData
     *
     */


    if ( params->tmpltSource == 2 )
    {
      if ( ! mddTmplt )
      {
        ABORT( status, CONDITIONDATAH_ENBNK, CONDITIONDATAH_MSGENBNK );
      }

      InspiralTmpltsFromILWD( status->statusPtr, params, mddTmplt );
      CHECKSTATUSPTR( status );
    }

    /* abort if we don't have a bank by now */
    if ( ! params->tmpltBankHead )
    {
      ABORT( status, CONDITIONDATAH_ENBNK, CONDITIONDATAH_MSGENBNK );
    }

#ifdef INSPIRAL_SO_DEBUG
    /* print out the template bank */
    PrintInspiralBank( status->statusPtr, params->tmpltBankHead, stdout ); 
    CHECKSTATUSPTR( status );
#endif


   /*
    *
    * create the linked list of template nodes for coarse templates
    *
    */


    /* make sure the linked list of template nodes is NULL */
    if ( params->masterParams.tmpltNodeHead )
    {
      ABORT( status, CONDITIONDATAH_ENNUL, CONDITIONDATAH_MSGENNUL );
    }

    /* we need to keep track of the number of templates for progress */
    params->masterParams.numTmpltsToFilter = 0;

    /* create the template node list of coarse templates */
    for ( currentTmplt = params->tmpltBankHead; currentTmplt; 
        currentTmplt = currentTmplt->next )
    { 
      LALFindChirpCreateTmpltNode( status->statusPtr, 
          currentTmplt, &tmpltNode );
      CHECKSTATUSPTR( status );

      /* save the head of the linked list */
      if ( ! params->masterParams.tmpltNodeHead ) 
        params->masterParams.tmpltNodeHead = tmpltNode;

      /* make sure we filter every segment against every coarse tmplt */
      for ( i = 0; i < tmpltNode->tmpltPtr->segmentIdVec->length; i++ )
        tmpltNode->tmpltPtr->segmentIdVec->data[i] = 1;

      params->masterParams.numTmpltsToFilter++;
    }

    /* record the total number of templates */
    params->masterParams.numTmpltsTotal = 
      params->masterParams.numTmpltsToFilter;

    /* if this is a bank simulation, set the number to filter to zero and */
    /* store the location of the bank head, so we can send the whole bank */
    if ( *(params->masterParams.inspiralDebugFlagPtr) == fcBankMinimalMatch )
    {
      params->masterParams.numTmpltsToFilter = 0;
      params->masterParams.tmpltBankHead = params->tmpltBankHead;
    }

    /* set the current template node to the head of the bank */
    params->masterParams.currentTmpltNode = 
      params->masterParams.tmpltNodeHead;

#ifdef INSPIRAL_SO_DEBUG
    /* print out the template node list */
    PrintTemplateNodeList( status->statusPtr, 
        params->masterParams.tmpltNodeHead, stdout );
    CHECKSTATUSPTR( status );

    fprintf( stdout, "numTmpltsToFilter = %d\n", 
        params->masterParams.numTmpltsToFilter );
    fflush( stdout );
#endif

  }
  else /* search slave */
  {


   /* 
    *
    * identify multiDimData and point the LAL structure pointers at the data
    *
    */


    for ( i = 0; i < inout->numberSequences; ++i )
    {
      if ( strstr( inout->sequences[i].name, INPUTNAME_CHANNEL ) )
      {


       /*
        *
        * data channel
        *
        */

        /* check that we are not getting multiple channels */
        if ( haveChan )
        {
          ABORT( status, CONDITIONDATAH_EMCHN, CONDITIONDATAH_MSGEMCHN );
        }

        /* check that the length of the data channel is correct */
        if ( inout->sequences[i].range.dTime.numberSamples != 
            params->inputDataLength )
        {
          ABORT( status, CONDITIONDATAH_ECHAN, CONDITIONDATAH_MSGECHAN );
        }

        /* check the units of the data channel */
        /* XXX not yet implemented as datacondAPI does not provide units XXX */

        if ( ! inout->sequences[i].range.dTime.channelName )
        {
          ABORT( status, CONDITIONDATAH_ENULL, 
              CONDITIONDATAH_MSGENULL " channelName" );
        }

        /* look for a known ifo and parse the channel name */
        {
          CHAR  knownIFONames[] = KNOWN_IFO_NAMES;
          CHAR *ifoNamePtr      = knownIFONames;

          memset( chanPtr->name, 0, LALNameLength * sizeof(CHAR) );
#ifdef INSPIRAL_SO_DEBUG
          fprintf( stdout, "channelName: %s\n", 
              inout->sequences[i].range.dTime.channelName[0] );
          fflush( stdout );
#endif

          /* loop over the list of known ifo names and copy the channel name */
          while ( *ifoNamePtr )
          {
#ifdef INSPIRAL_SO_DEBUG
            fprintf( stdout, "searching for %c%c\n",
                *ifoNamePtr, *(ifoNamePtr+1) );
            fflush( stdout );
#endif

            if ( ! strncmp( inout->sequences[i].range.dTime.channelName[0],
                  ifoNamePtr, 2 * sizeof(CHAR) ) )
            {
              strncpy( chanPtr->name, 
                  inout->sequences[i].range.dTime.channelName[0],
                  LALNameLength * sizeof(CHAR) );
              memcpy( ifo, inout->sequences[i].range.dTime.channelName[0],
                  2 * sizeof(CHAR) );
              ifo[2] = '\0';
#ifdef INSPIRAL_SO_DEBUG
              fprintf( stdout, "ifo: %s\n", ifo );
              fprintf( stdout, "channel: %s\n", chanPtr->name );
              fflush( stdout );
#endif
              break;
            }
            ifoNamePtr += 2;
          }
        }

        /* if we didn't find a known ifo name, give up */
        if ( ! *(chanPtr->name) )
        {
          ABORT( status, CONDITIONDATAH_EUIFO, CONDITIONDATAH_MSGEUIFO );
        }

        /* set the rest of the parameters of the data channel time series */
        params->info.inputStartTime.gpsSeconds =
          chanPtr->epoch.gpsSeconds = 
          inout->sequences[i].range.dTime.startSec;
        params->info.inputStartTime.gpsNanoSeconds =
          chanPtr->epoch.gpsNanoSeconds = 
          inout->sequences[i].range.dTime.startNan;
        chanPtr->deltaT = inout->sequences[i].range.dTime.timeStepSize;
        chanPtr->f0 = inout->sequences[i].range.dTime.baseFreq;
        chanPtr->sampleUnits = lalADCCountUnit;

        /* save the channel end time incase we are doing injections */
        params->info.inputEndTime.gpsSeconds =
          chanStopEpoch.gpsSeconds = 
          inout->sequences[i].range.dTime.stopSec;
        params->info.inputEndTime.gpsNanoSeconds =
          chanStopEpoch.gpsNanoSeconds = 
          inout->sequences[i].range.dTime.stopNan;
        LALGPStoINT8( status->statusPtr, &chanStopTime, &chanStopEpoch );
        CHECKSTATUSPTR( status );

        /* point (or copy if simulation) the actual data channel */
        chanPtr->data->length = 
          params->slaveParams.dataChannel.length =
          inout->sequences[i].range.dTime.numberSamples;
        chanPtr->data->data   = 
          params->slaveParams.dataChannel.data =
          inout->sequences[i].data.real4;

        /* set sample rate for findchirp functions */
        params->slaveParams.dataParams->deltaT =
          params->slaveParams.tmpltParams->deltaT = 
          params->slaveParams.filterParams->deltaT = 
          inout->sequences[i].range.dTime.timeStepSize;

        /* store the data start time in nanoseconds */
        LALGPStoINT8( status->statusPtr, &chanStartTime, &(chanPtr->epoch) );
        CHECKSTATUSPTR( status );

        /* store a pointer to the start of the data segment */
        dataPtr = chanPtr->data->data;

        /* we have the data channel */
        haveChan += 1;
      }
      /* to ensure that the spectrum gets picked up correctly */
      else if ( strstr( inout->sequences[i].name, INPUTNAME_SPECTRUM ) &&
           strstr( inout->sequences[i].name, INPUTNAME_PSD ) )
      {


       /*
        *
        * power spectrum
        *
        */

        if ( haveSpec )
        {
          ABORT( status, CONDITIONDATAH_EMPSD, CONDITIONDATAH_MSGEMPSD );
        }

        /* check that we are supposed to be getting a spectrum from the  */
        /* datacondAPI and abort if not                                  */
        if ( params->slaveParams.specType )
        {
          ABORT( status, CONDITIONDATAH_EDCSP, CONDITIONDATAH_MSGEDCSP );
        }

        /* check that the length of the power spectrum is correct */
        if ( inout->sequences[i].range.dFreq.numberSamples != 
            params->initParams.numPoints / 2 + 1 )
        {
          ABORT( status, CONDITIONDATAH_ESPEC, CONDITIONDATAH_MSGESPEC );
        }

        /* check the units of the power spectrum */
        /* XXX not yet implemented as datacondAPI does not provide units XXX */

        /* set the parameters of the power spectrum frequency series */
        strncpy( specPtr->name, inout->sequences[i].name, LALNameLength );
        specPtr->epoch.gpsSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeSec;
        specPtr->epoch.gpsNanoSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeNan;
        specPtr->f0 = inout->sequences[i].range.dFreq.baseFreq;
        specPtr->deltaF = inout->sequences[i].range.dFreq.freqStepSize;
        {
          RAT4 negRootTwo = { -1, 1 };
          LALUnit unit;
          LALUnitPair pair;

          specPtr->sampleUnits = lalADCCountUnit;
          pair.unitOne = &lalADCCountUnit;
          pair.unitTwo = &lalHertzUnit;
          TRY( LALUnitRaise( status->statusPtr, &unit, 
                pair.unitTwo, &negRootTwo ), status );
          pair.unitTwo = &unit;
          TRY( LALUnitMultiply( status->statusPtr, &(specPtr->sampleUnits),
                &pair ), status );
        }

        /* point at the actual power spectrum */
        specPtr->data->length = inout->sequences[i].range.dFreq.numberSamples;
        specPtr->data->data   = inout->sequences[i].data.real4;

        /* record that we have a psd */
        haveSpec += params->slaveParams.haveSpec = 1;

      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_SPECTRUM ) && 
          strstr( inout->sequences[i].name, INPUTNAME_RESPONSE ) )
      {


       /*
        *
        * response function
        *
        */

        if ( haveResp )
        {
          ABORT( status, CONDITIONDATAH_EMRSP, CONDITIONDATAH_MSGEMRSP );
        }

        /* check that the length of the response function is correct */
        if ( inout->sequences[i].range.dFreq.numberSamples != 
            params->initParams.numPoints / 2 + 1 )
        {
          ABORT( status, CONDITIONDATAH_ERESP, CONDITIONDATAH_MSGERESP );
        }

        /* check the units of the response function */
        /* XXX not yet implemented as datacondAPI does not provide units XXX */

        /* set the parameters of the response function frequency series */
        strncpy( respPtr->name, INPUTNAME_RESPONSE, 
            LALNameLength * sizeof(CHAR) );
        respPtr->epoch.gpsSeconds = 
          params->responseEpoch.gpsSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeSec;
        respPtr->epoch.gpsNanoSeconds = 
          params->responseEpoch.gpsNanoSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeNan;
        respPtr->f0 = inout->sequences[i].range.dFreq.baseFreq;
        respPtr->deltaF = inout->sequences[i].range.dFreq.freqStepSize;
        respPtr->sampleUnits = strainPerCount;

        /* point at the actual response function data */
        respPtr->data->length = inout->sequences[i].range.dFreq.numberSamples;
        respPtr->data->data   = inout->sequences[i].data.complex8;

        /* record that we have an old type static calibration */
        params->responseType = oldCalibration;
        
        haveResp += 1;
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_TMPLTBANK ) )
      {
        /* we do nothing with the template bank on the slaves, but we     */
        /* have to test for it, otherwise we can't abort if there is an   */
        /* unknown channel name                                           */
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_CAL ) )
      {
        /* the callibration should be processes by the extract            */
        /* response function                                              */
        haveNewCal = 1;
        if ( strstr( inout->sequences[i].name, "RESPONSE" ) )
        {
          if ( haveResp )
          {
            ABORT( status, CONDITIONDATAH_EMRSP, CONDITIONDATAH_MSGEMRSP );
          }

          calIdx = i;

          responseEpoch.gpsSeconds = 
            inout->sequences[i].range.dFreq.gpsStartTimeSec;
          responseEpoch.gpsNanoSeconds = 
            inout->sequences[i].range.dFreq.gpsStartTimeNan;

          haveResp += 1;
        }
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_INJEPOCHS ) )
      {
        if ( params->slaveParams.simParams )
        {
          mddInjEpochs = &( inout->sequences[i] );
        }
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_INJPARAMS ) )
      {
        if ( params->slaveParams.simParams )
        {
          mddInjParams = &( inout->sequences[i] );
        }
      }
      else
      {
        ABORT( status, CONDITIONDATAH_EINPT, CONDITIONDATAH_MSGEINPT ); 
      }
    }


   /*
    *
    * check that we have everything that we need to continue
    *
    */


    if ( ! haveChan )
    {
      ABORT( status, CONDITIONDATAH_ENCHN, CONDITIONDATAH_MSGENCHN );
    }

    if ( ! haveSpec && ! params->slaveParams.specType )
    {
      ABORT( status, CONDITIONDATAH_ENPSD, CONDITIONDATAH_MSGENPSD );
    }

    if ( ! haveResp )
    {
      ABORT( status, CONDITIONDATAH_ENRSP, CONDITIONDATAH_MSGENRSP );
    }


   /*
    *
    * read in the new type of calibration if it is present
    *
    */


    /* check and see if we have a new type calibration in the input */
    if ( haveNewCal )
    {
      /* set the start epoch for the response function from the data */
      respPtr->epoch.gpsSeconds = chanPtr->epoch.gpsSeconds;
      respPtr->epoch.gpsNanoSeconds = chanPtr->epoch.gpsNanoSeconds;

      /* set the frequency step size of the response from the psd */
      respPtr->f0 = 0.0;
      respPtr->deltaF = chanPtr->deltaT / 
        ( (REAL8) params->initParams.numPoints );

      /* set the units of the desired response function */
      respPtr->sampleUnits = strainPerCount;

      /* allocate memory for the vector in the response function */
      respPtr->data->length = params->initParams.numPoints / 2 + 1;
      respPtr->data->data = (COMPLEX8 *) 
        LALCalloc( respPtr->data->length, sizeof(COMPLEX8) );

      /* set the name of the response function and extract it */
      strncpy( respPtr->name, INPUTNAME_RESPONSE, 
          LALNameLength * sizeof(CHAR) );
      LALExtractResponse( status->statusPtr, respPtr, inout, ifo );
      CHECKSTATUSPTR( status );
#if 0
      if ( status->statusPtr->statusCode == -1 &&
          ( status->statusPtr->statusPtr->statusCode == CALIBRATIONH_EZERO ||
            status->statusPtr->statusPtr->statusCode == CALIBRATIONH_ETIME ) )
      {
        DETATCHSTATUSPTR( status );
        ATTATCHSTATUSPTR( status );

        /* free the memory that was allocated for the new calibration */
        LALFree( respPtr->data->data );

        /* check that the length of the response function is correct */
        if ( inout->sequences[calIdx].range.dFreq.numberSamples != 
            params->initParams.numPoints / 2 + 1 )
        {
          ABORT( status, CONDITIONDATAH_ERESP, CONDITIONDATAH_MSGERESP );
        }

        /* set the parameters of the response function frequency series */
        respPtr->epoch.gpsSeconds = 
          params->responseEpoch.gpsSeconds = 
          inout->sequences[calIdx].range.dFreq.gpsStartTimeSec;
        respPtr->epoch.gpsNanoSeconds = 
          params->responseEpoch.gpsNanoSeconds = 
          inout->sequences[calIdx].range.dFreq.gpsStartTimeNan;
        respPtr->f0 = inout->sequences[calIdx].range.dFreq.baseFreq;
        respPtr->deltaF = inout->sequences[calIdx].range.dFreq.freqStepSize;
        respPtr->sampleUnits = strainPerCount;

        /* point at the actual response function data */
        respPtr->data->length = 
          inout->sequences[calIdx].range.dFreq.numberSamples;
        respPtr->data->data   = 
          inout->sequences[calIdx].data.complex8;

        /* record that we have an old type static calibration */
        params->responseType = oldCalibration;
      }
      else
      {
        CHECKSTATUSPTR( status );

        /* record what type and time of response function we are using */
        params->responseType = newCalibration;
        memcpy( &(params->responseEpoch), &responseEpoch, sizeof(LIGOTimeGPS) );
      }
#endif
        params->responseType = newCalibration;
        memcpy( &(params->responseEpoch), &responseEpoch, sizeof(LIGOTimeGPS) );
    }


   /*
    *
    * make sure that memory for the PSD exists
    *
    */


    if ( params->slaveParams.specType )
    {
      if ( params->slaveParams.haveSpec )
      {
        /* should never get here as we have already aborted */
        ABORT( status, CONDITIONDATAH_EALOC, CONDITIONDATAH_MSGEALOC );
      }

      /* allocate memory for the power spectrum to be generated later */
      specPtr->data->length = params->initParams.numPoints / 2 + 1;
      if ( ! (specPtr->data->data   = (REAL4 *)
            LALCalloc( specPtr->data->length, sizeof(REAL4) )) )
      {
        ABORT( status, CONDITIONDATAH_EALOC, CONDITIONDATAH_MSGEALOC );
      }
    }


   /* 
    *
    * create each individual DataSegment in the DataSegmentVector
    *
    */


    for ( i = 0; i < params->dataSegVec->length; ++i )
    {
      /* point to current segment */
      DataSegment      *currentSegment = params->dataSegVec->data + i;
      REAL4Sequence    *origDataPtr;

      /* this should be set to a unique number for each segment   */
      currentSegment->number = i;

      /* copy the REAL4TimeSeries information into the data segment */
      origDataPtr = currentSegment->chan->data;
      memcpy( currentSegment->chan, chanPtr, sizeof(REAL4TimeSeries) );
      currentSegment->chan->data = origDataPtr;

      /* set the correct GPS time for the current data segment */
      {
        INT8 currentSegmentTime = chanStartTime + (INT8) 
          ( 1.0e9 * 
            ((REAL8) params->initParams.numPoints - (REAL8) params->ovrlap) *
            (REAL8) i * chanPtr->deltaT
          );

        LALINT8toGPS( status->statusPtr, &(currentSegment->chan->epoch), 
            &currentSegmentTime );
        CHECKSTATUSPTR( status );
      }

      /* set up the REAL4Sequence to contain the correct data */
      currentSegment->chan->data->length = params->initParams.numPoints;
      currentSegment->chan->data->data = dataPtr;

      /* advance the dataPtr to the next segment */
      dataPtr +=  currentSegment->chan->data->length - params->ovrlap;

      /* power spectrum */
      currentSegment->spec = &(params->spec);

      /* response function */
      currentSegment->resp = &(params->resp);
    }


    /*
     *
     * perform setup for simulations
     *
     */


    if ( params->slaveParams.simParams )
    {
      if ( params->slaveParams.simParams->simType == fcBankMinimalMatch ) 
      {
#ifdef INSPIRAL_USE_URAND
        /* use the kernel random number generator to get a seed */
        FILE   *fpRand = NULL;
        INT4    randByte;

        params->seed = 0;

        fpRand = fopen( "/dev/urandom", "r" );

        if ( fpRand )
        {
          for ( randByte = 0; randByte < 4 ; ++randByte )
          {
            INT4 tmpSeed = (INT4) fgetc( fpRand );
            params->seed += tmpSeed << ( randByte * 8 );
          }
          fclose( fpRand );
        }
        else
        {
          ABORT( status, 999, "cannot open /dev/urandom" );
        }
#endif
      }

      /* create the random parameter structure */
      LALCreateRandomParams( status->statusPtr, 
          &(params->slaveParams.simParams->randomParams), params->seed );
      CHECKSTATUSPTR( status );

      if ( params->slaveParams.simParams->simType == fcRealDataInject ||
          params->slaveParams.simParams->simType == fcGaussianNoiseInject  ) 
      {
        InspiralInjection      *prevEvent = NULL;

        /* parse the multiDimData containing the injection parameters */
        UINT4     numEpochs = mddInjEpochs->dimensions[1];
        UINT4     numParams = mddInjParams->dimensions[1];
        UINT4     paramIndx = mddInjParams->dimensions[0];
        REAL4    *injParams = mddInjParams->data.real4;;

        /* XXX sanity checks missing XXX */
        if ( numEpochs != numParams )
        {
          ABORT( status, 999, "numepochs not equal to numparams" );
        }

        /* step through the list of injection times and create an         */
        /* InspiralInjection structure for each time that lies within the */
        /* current data chunk                                             */
        for ( i = 0; i < numEpochs; ++i )
        {
          InspiralInjection      *thisEvent = NULL;
          INT8                    injectTime;
          LIGOTimeGPS             injectEpoch;

          /* get the time of the injection */
          injectEpoch.gpsSeconds     = mddInjEpochs->data.int4s[2*i];
          injectEpoch.gpsNanoSeconds = mddInjEpochs->data.int4s[2*i+1];

          LALGPStoINT8( status->statusPtr, &injectTime, &injectEpoch );
          CHECKSTATUSPTR( status );

          if ( injectTime > chanStartTime && injectTime <= chanStopTime )
          {
            if ( ! params->slaveParams.simParams->injectEvent )
            {
              params->slaveParams.simParams->injectEvent = thisEvent =
                (InspiralInjection *) LALCalloc( 1, sizeof(InspiralInjection) );
            }
            else
            {
              prevEvent->next = thisEvent = 
                (InspiralInjection *) LALCalloc( 1, sizeof(InspiralInjection) );
            }

            thisEvent->coaTime = injectTime;

            /* read the parameters for the event */
            thisEvent->totalMass    = injParams[paramIndx*i];
            thisEvent->eta          = injParams[paramIndx*i+1];
            thisEvent->dist         = injParams[paramIndx*i+2];
            thisEvent->inclination  = injParams[paramIndx*i+3];
            thisEvent->coaPhase     = injParams[paramIndx*i+4];
            thisEvent->longitude    = injParams[paramIndx*i+5];
            thisEvent->latitude     = injParams[paramIndx*i+6];
            thisEvent->polarization = injParams[paramIndx*i+7];

            prevEvent = thisEvent;
          }
        }
      }
    }


    /* 
     *
     * if this is the first slave create a standard candle 
     *
     */

    
    if ( commRank == 1 )
    {
      FindChirpStandardCandle   *candle;
      candle = params->slaveParams.candlePtr = (FindChirpStandardCandle *)
        LALCalloc( 1, sizeof(FindChirpStandardCandle) );

      strncpy( candle->ifo, params->chan.name, 2 * sizeof(CHAR) );
      candle->tmplt.mass1 = CANDLE_MASS1;
      candle->tmplt.mass2 = CANDLE_MASS2;
      candle->rhosq = CANDLE_RHOSQ;

      candle->tmplt.totalMass = candle->tmplt.mass1 + candle->tmplt.mass2;
      candle->tmplt.mu = candle->tmplt.mass1 * candle->tmplt.mass2 / 
        candle->tmplt.totalMass;
      candle->tmplt.eta = candle->tmplt.mu / candle->tmplt.totalMass;
    }
  }


  /* normal exit */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}


#ifdef INSPIRAL_SO_DEBUG
/*-----------------------------------------------------------------------*/
static void
PrintInspiralBank (
    LALStatus           *status,
    InspiralTemplate    *head,
    FILE                *fp
    )
{
  InspiralTemplate    *current;

  INITSTATUS( status, "PrintInspiralBank", CONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  ASSERT( head, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );


 /*
  *
  * print the inspiral template parameter bank
  *
  */


  for ( current = head; current; current = current->next)
  {
    fprintf( fp, 
        "%p\t%d\t%f\t%f\t%f\t%f\t%f\t%p\t%p\n",
        current, current->number, 
        current->mass1, current->mass2, current->totalMass,
        current->eta, current->mu,
        current->next, current->fine );
    fflush( fp );
    if ( current->fine )
    {
      PrintInspiralBank( status->statusPtr, current->fine, fp );
      CHECKSTATUSPTR( status );
    }
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*-----------------------------------------------------------------------*/
static void
PrintTemplateNodeList(
    LALStatus                  *status,
    InspiralTemplateNode       *tmpltHead,
    FILE                       *fp
    )
{
  InspiralTemplateNode *current = NULL;

  INITSTATUS( status, "PrintTemplateNodeList", CONDITIONDATAC);

  ASSERT( tmpltHead, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );

  for ( current = tmpltHead; current; current = current->next )
  {
    fprintf( fp, 
        "-----------------------------------------------------------------\n" );
    fprintf( fp, "Template Node at %p points to template at %p\n",
        current, current->tmpltPtr );
    fflush( fp );
    fprintf( fp, "   number %d: m1 = %f, m2 = %f\n",
        current->tmpltPtr->number, 
        current->tmpltPtr->mass1, 
        current->tmpltPtr->mass2 );
    fflush( fp );
    fprintf( fp, "next node: %p, prev node: %p\n",
        current->next, current->prev );
    fflush( fp );
  }

  RETURN( status );
}
#endif
